import pytest
from numpy.testing import assert_allclose
from test_data.element_results import (
    ELEMENT_HEX_TYPES,
    ELEMENT_QUAD_TYPES,
    ELEMENT_TRI_TYPES,
)

from pybaqus.reader import open_fil, read_records


@pytest.mark.parametrize("test_case", ELEMENT_QUAD_TYPES, ids=lambda x: x["name"])
def test_element_quad(test_case):
    """Test element stress results for different element types.

    Parameters
    ----------
        test_case:
            Dictionary containing test parameters and expected results
    """
    model = open_fil(f"tests/abaqus/fil/quad_{test_case['name']}.fil")
    stress = model.get_nodal_result(var=test_case["variable"], step=1, inc=1)
    assert_allclose(
        stress,
        test_case["expected"],
        err_msg=f"Failed for element type: {test_case['name']}",
        rtol=1e-6,
    )


@pytest.mark.parametrize("test_case", ELEMENT_TRI_TYPES, ids=lambda x: x["name"])
def test_element_tri(test_case):
    """Test element stress results for different element types.

    Parameters
    ----------
        test_case:
            Dictionary containing test parameters and expected results
    """
    model = open_fil(f"tests/abaqus/fil/tri_{test_case['name']}.fil")
    stress = model.get_nodal_result(var=test_case["variable"], step=1, inc=1)
    assert_allclose(
        stress,
        test_case["expected"],
        err_msg=f"Failed for element type: {test_case['name']}",
        rtol=1e-6,
    )


@pytest.mark.parametrize("test_case", ELEMENT_HEX_TYPES, ids=lambda x: x["name"])
def test_element_hex(test_case):
    """Test element stress results for different element types.

    Parameters
    ----------
        test_case:
            Dictionary containing test parameters and expected results
    """
    model = open_fil(f"tests/abaqus/fil/hex_{test_case['name']}.fil")
    stress = model.get_nodal_result(var=test_case["variable"], step=1, inc=1)
    assert_allclose(
        stress,
        test_case["expected"],
        err_msg=f"Failed for element type: {test_case['name']}",
        rtol=1e-6,
    )



import pytest
from numpy.testing import assert_allclose

from pybaqus.reader import open_fil, read_records


@pytest.fixture
def fil_path_2d():
    return "tests/abaqus/fil/quad_CPE4.fil"


@pytest.fixture
def fil_path_3d():
    return "tests/abaqus/fil/hex_C3D8.fil"


def test_read_records(fil_path_2d):
    records = read_records(fil_path_2d)
    assert hasattr(records, "__iter__")  # Check if it's an iterator


@pytest.mark.parametrize("progress", [False, True])
def test_open_fil(fil_path_2d, progress, capsys):
    result = open_fil(fil_path_2d, progress=progress)
    assert result is not None

    captured = capsys.readouterr()
    if progress:
        assert "Reading records..." in captured.out
        assert "Parsing records..." in captured.out
    else:
        assert captured.out == ""


def test_open_fil_nonexistent_file():
    with pytest.raises(FileNotFoundError):
        open_fil("nonexistent_file.fil")


def test_abaqus_release(fil_path_2d):
    model = open_fil(fil_path_2d)
    # Check the release
    assert model.release["release"] == "6.23-1"


def test_heading(fil_path_2d):
    model = open_fil(fil_path_2d)
    assert model.heading == "Test elements of the type CPE4 with quad shape"


def test_model_size(fil_path_2d):
    model = open_fil(fil_path_2d)
    assert model.size["elements"] == len(model.elements)
    assert model.size["nodes"] == len(model.nodes)


def test_discontinuous_nodes():
    model = open_fil("tests/abaqus/fil/discontinuous_numbering_2D.fil")
    assert model.size["nodes"] == len(model.nodes)


def test_parse_node_2d(fil_path_2d):
    model = open_fil(fil_path_2d)
    n1 = model.nodes[1]
    n2 = model.nodes[2]
    n3 = model.nodes[3]
    n4 = model.nodes[4]

    assert_allclose(n1.coords, [0.1, 0.2, 0.0])
    assert_allclose(n2.coords, [12.9, 0.2, 0.0])
    assert_allclose(n3.coords, [0.1, 10.5, 0.0])
    assert_allclose(n4.coords, [12.9, 10.5, 0.0])


def test_parse_node_3d(fil_path_3d):
    model = open_fil(fil_path_3d)
    n1 = model.nodes[1]
    n2 = model.nodes[2]
    n3 = model.nodes[3]
    n4 = model.nodes[4]
    n5 = model.nodes[5]
    n6 = model.nodes[6]
    n7 = model.nodes[7]
    n8 = model.nodes[8]

    assert_allclose(n1.coords, [0.0, 0.0, 0.0])
    assert_allclose(n2.coords, [10.0, 0.0, 0.0])
    assert_allclose(n3.coords, [0.0, 20.0, 0.0])
    assert_allclose(n4.coords, [10.0, 20.0, 0.0])
    assert_allclose(n5.coords, [0.0, 0.0, 30.0])
    assert_allclose(n6.coords, [10.0, 0.0, 30.0])
    assert_allclose(n7.coords, [0.0, 20.0, 30.0])
    assert_allclose(n8.coords, [10.0, 20.0, 30.0])


@pytest.mark.parametrize(
    "file_path,expected_record_count",
    [
        ("tests/abaqus/fil/quad_CPE4.fil", 50),
    ],
)
def test_read_records_count(file_path, expected_record_count):
    records = list(read_records(file_path))
    assert len(records) == expected_record_count



from pybaqus import open_fil
import pyvista as pv

res = open_fil("model_results.fil")

mesh = res.get_deformed_mesh(step=1, inc=1, scale=3)
U1 = res.get_nodal_result(var="U1", step=1, inc=1)
mesh.point_arrays["U1"] = U1


plot = pv.Plotter()
plot.add_mesh(mesh, show_edges=True, scalars="U1", show_scalar_bar=True)
plot.view_xy()
plot.show()



import os

from abaqus import *
import assembly
import datum
import interaction
import job
import load
import material
import mesh
import part
import partition
import regionToolset
import section
import sketch
import step
from abaqus import backwardCompatibility, mdb, monitorManager, session
from abaqusConstants import *
from jobMessage import *
from odbAccess import Odb

# Set backward compatibility
backwardCompatibility.setValues(includeDeprecated=False, reportDeprecated=True)


def get_ascii_job(job_obj):
    """Configure Job to save results in the ASCII format (*.fil).

    Parameters
    ----------
    job_obj : TODO

    """
    job_obj.writeInput(consistencyChecking=False)

    input_lines = (
        "*FILE FORMAT, ASCII\n"
        + "*EL FILE\n"
        + "S, E, COORD\n"
        + "*NODE FILE\n"
        + "COORD, U\n"
    )

    # Open *inp file
    file_name = job_obj.name + ".inp"

    with open(file_name, "r") as inp_file:
        lines = inp_file.readlines()

    with open(file_name, "w") as inp_file:
        for l in lines:
            if l == "*End Step\n":
                inp_file.writelines(input_lines)
                inp_file.write(l)
            else:
                inp_file.write(l)

    job_ascii = mdb.JobFromInputFile(
        name=job_obj.name,
        inputFileName=job_obj.name + ".inp",
        type=ANALYSIS,
        memory=90,
        memoryUnits=PERCENTAGE,
        getMemoryFromAnalysis=True,
        explicitPrecision=SINGLE,
        nodalOutputPrecision=SINGLE,
        userSubroutine=job_obj.userSubroutine,
        scratch=job_obj.scratch,
        multiprocessingMode=DEFAULT,
        numCpus=job_obj.numCpus,
        numDomains=job_obj.numDomains,
    )

    return job_ascii


# Define some parameters
x0 = 0.1
y0 = 0.2
w = 12.8
h = 10.3

# Model
model = mdb.Model(name="test")
# Viewport
viewp = session.Viewport(name="test", origin=(20, 0), width=450, height=170)

sketch_test = model.ConstrainedSketch(name="test", sheetSize=200.0)
sketch_test.rectangle(point1=(x0, y0), point2=(x0 + w, y0 + h))

test_part = model.Part(name="test", dimensionality=TWO_D_PLANAR, type=DEFORMABLE_BODY)
test_part.BaseShell(sketch=sketch_test)

props = (1000, 0.3)

test_mat = model.Material(name="test-material")
test_mat.Elastic(table=(props,), type=ISOTROPIC)

test_sec = model.HomogeneousSolidSection(
    name="test-section", material=test_mat.name, thickness=1
)
# Create sets to assign section to
set_faces = test_part.Set(faces=test_part.faces[:], name="set-test_part")

test_part.SectionAssignment(
    region=set_faces, sectionName="test-section", thicknessAssignment=FROM_SECTION
)

asmbly = model.rootAssembly
inst = asmbly.Instance(name="name", part=test_part, dependent=False)

step_test = model.StaticStep(
    name="test-step",
    previous="Initial",
    nlgeom=False,
    timeIncrementationMethod=AUTOMATIC,
    initialInc=1,
    maxInc=1.0,
    maxNumInc=10,
)

# Load definition
vertex_load = inst.vertices.findAt(((x0 + w, y0 + h, 0),))
region_shell = regionToolset.Region(vertices=vertex_load)
model.ConcentratedForce(
    name="test-load", createStepName=step_test.name, region=region_shell, cf2=1
)

# Boundary conditions
vert_1 = inst.vertices.findAt(((x0, y0, 0),))
vert_2 = inst.vertices.findAt(((x0 + w, y0, 0),))
region_bottom_1 = regionToolset.Region(vertices=vert_1)
region_bottom_2 = regionToolset.Region(vertices=vert_2)
model.DisplacementBC(
    name="test-bc-1", createStepName="Initial", region=region_bottom_1, u1=0, u2=0
)
model.DisplacementBC(
    name="test-bc-2", createStepName="Initial", region=region_bottom_2, u2=0
)

# Mesh the model
mesh_region = (inst.faces,)
elemType_1 = mesh.ElemType(elemCode=CPS4R, elemLibrary=STANDARD)

asmbly.setElementType(regions=mesh_region, elemTypes=(elemType_1,))
asmbly.seedPartInstance(regions=(inst,), size=(w))
asmbly.generateMesh(regions=(inst,))

job_test = mdb.Job(
    name="test",
    model=model.name,
    description="some description",
    numCpus=1,
    numDomains=1,
)

get_ascii_job(job_obj=job_test)



import pytest
from pybaqus.model import Model
from pybaqus.nodes import Node2D, Node3D

@pytest.fixture
def empty_model():
    return Model()

# @pytest.fixture
# def new_2D_node():
#     return Node2D(1, dof_map=[])

def test_model_initialization(empty_model):
    assert isinstance(empty_model, Model)
    assert len(empty_model.nodes) == 0
    assert len(empty_model.elements) == 0
    assert len(empty_model.element_sets) == 0
    assert len(empty_model.node_sets) == 0
    assert len(empty_model.surfaces) == 0

def test_add_node(empty_model):
    class MockNode:
        def __init__(self, num):
            self._num = num

    node = MockNode(1)
    empty_model.add_node(node)
    assert len(empty_model.nodes) == 1
    assert empty_model.nodes[1] == node

def test_add_element(empty_model):
    class MockElement:
        def __init__(self, num):
            self.num = num

    element = MockElement(1)
    empty_model.add_element(element)
    assert len(empty_model.elements) == 1
    assert empty_model.elements[1] == element

def test_add_set(empty_model):
    elements = [1, 2, 3]
    result = empty_model.add_set("test_set", elements, "element")
    assert "test_set" in empty_model.element_sets
    assert empty_model.element_sets["test_set"] == elements
    assert result == elements

    nodes = [4, 5, 6]
    result = empty_model.add_set("test_node_set", nodes, "node")
    assert "test_node_set" in empty_model.node_sets
    assert empty_model.node_sets["test_node_set"] == nodes
    assert result == nodes




"""
This file contains a utility function to generate and modify the Abaqus
input file (*.inp) so that the results are written in the ASCII output
format (*.fil).

Author: Cristóbal Tapia Camú
E-mail: crtapia@gmail.com
"""


def get_ascii_job(job_obj):
    """Configure Job to save results in the ASCII format (*.fil).

    Parameters
    ----------
    job_obj : TODO

    """
    job_obj.writeInput(consistencyChecking=False)

    input_lines = ("*FILE FORMAT, ASCII\n" + "*EL FILE\n" + "S, E, COORD\n" +
                   "*NODE FILE\n" + "COORD, U\n")

    # Open *inp file
    file_name = job_obj.name + ".inp"

    with open(file_name, "r") as inp_file:
        lines = inp_file.readlines()

    with open(file_name, "w") as inp_file:
        for l in lines:
            if l == "*End Step\n":
                inp_file.writelines(input_lines)
                inp_file.write(l)
            else:
                inp_file.write(l)

    job_ascii = mdb.JobFromInputFile(name=job_obj.name, inputFileName=job_obj.name + ".inp",
                                     type=ANALYSIS, memory=90, memoryUnits=PERCENTAGE,
                                     getMemoryFromAnalysis=True, explicitPrecision=SINGLE,
                                     nodalOutputPrecision=SINGLE,
                                     userSubroutine=job_obj.userSubroutine,
                                     scratch=job_obj.scratch, multiprocessingMode=DEFAULT,
                                     numCpus=job_obj.numCpus, numDomains=job_obj.numDomains)

    return job_ascii



"""
Utilities to read the ASCII *.fil files generates by abaqus.
"""

import re

from pybaqus.fil_result import FilParser


def open_fil(file_name, progress=False):
    """Read the *.fil file

    Parameters
    ----------
    file_name : str
        Path to the *.fil file
    progress : bool
        Indicates whether the progress of the reading process should be shown
        in a progress bar. (default: False)

    Returns
    -------
    Result : Object containing the results of the *.fil file

    """
    try:
        if progress:
            print("Reading records...")

        records = read_records(file_name)

        # Create result object
        if progress:
            print("Parsing records...")

        result = FilParser(records, progress=progress)
        del records

        return result.model

    except FileNotFoundError as e:
        raise e
    except Exception as e:
        raise e


def read_records(file_name: str):
    with open(file_name, "r") as result:
        content = result.read().replace("\n", "")
        # Make a list with each record.
        # Each record starts with `*`, so we just need to use a regex to return each
        # record.
        pattern = r"(?<=\*)+.+?(?=\*)|(?<=\*)+.+?(?=$)"
        records = re.findall(pattern, content)
        del content

    return records



"""
Definition of node objects.
"""

import numpy as np


class Node:
    """Define a node in the finite element model.

    Parameters
    ----------
    x : float
    y : float
    z : float
    num : float

    """

    def __init__(self, num: int, model):
        self._num: int = num
        self.model = model
        self._in_elements = None

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @property
    def rx(self):
        return self._rx

    @property
    def ry(self):
        return self._ry

    @property
    def rz(self):
        return self._rz

    @property
    def coords(self):
        return self._get_coords()

    @property
    def in_elements(self):
        return self._in_elements

    @in_elements.setter
    def in_elements(self, x):
        self._in_elements = x

    def _get_coords(self):
        print("This method needs to be redefined in subclass")
        return 1


class Node2D(Node):
    """Two-dimensional node.

    Parameters
    ----------
    num : int
        ID of the node
    dof_map : dict
        Dictionary mapping the active DOF to each nodal output position
    model : `obj`:Model
        Model to which the node belongs to.
    *dof :
        The values for all degrees of freedom.

    """
    _x: float
    _y: float
    _rz: float

    def __init__(self, num: int, dof_map, model, *dof):
        super().__init__(num, model)

        self._x = dof[dof_map[0] - 1]
        self._y = dof[dof_map[1] - 1]
        self._rz = dof[dof_map[5] - 1] if dof_map[5] > 0 else np.nan
        self._num = num

    def _get_coords(self):
        return (self._x, self._y, 0)


class Node3D(Node):
    """Three-dimensional node.

    Parameters
    ----------
    num : int
        Number of the node
    dof_map : dict
        Dictionary mapping the active DOF to each nodal output position
    model : `obj`:Model
        Model to which the node belongs to.
    *dof :
        The values for all degree of freedom.

    """
    _x: float
    _y: float
    _z: float
    _rx: float
    _ry: float
    _rz: float

    def __init__(self, num, dof_map, model, *dof):
        super().__init__(num, model)

        self._x = dof[dof_map[0] - 1]
        self._y = dof[dof_map[1] - 1]
        self._z = dof[dof_map[2] - 1]
        self._rx = dof[dof_map[3] - 1] if dof_map[3] > 0 else np.nan
        self._ry = dof[dof_map[4] - 1] if dof_map[4] > 0 else np.nan
        self._rz = dof[dof_map[5] - 1] if dof_map[5] > 0 else np.nan
        self._num = num

    def _get_coords(self):
        return (self._x, self._y, self._z)
